# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import discord

from detache.command import CommandInherit
from detache.wrappers import EventListenerInherit, BgTaskInherit


class Plugin:
    """
    Plugin class. Create your own plugins by inheriting this class.
    """

    __plugin_name__ = "Plugin"

    def __init__(self, bot):
        #: Bot the plugin belongs to
        self.bot = bot

        #: aiohttp.ClientSession to be used for any HTTP requests
        self.http = self.bot.http_session

        self.log = self.bot.log

        self.commands = self.find_commands()
        self.event_listeners = self.find_event_listeners()
        self.bg_tasks = self.find_bg_tasks()

    def __repr__(self):
        return "Plugin({!r})".format(self.__plugin_name__)

    def find_commands(self):
        """Returns dict of commands."""

        self.log.debug("finding commands in {!r}".format(self))

        commands = {}

        for name in dir(self):
            o = getattr(self, name)
            if issubclass(o.__class__, CommandInherit):  # check for command objects
                o.plugin = self
                commands[o.name] = o

                self.log.debug("found command: {!r}".format(o.name))

        return commands

    def find_event_listeners(self):
        """Returns dict of event listeners."""

        self.log.debug("finding event listeners in {!r}".format(self))

        listeners = {}
        # example:
        # {
        #     "on_message": [<func>, <func>],
        #     "on_message_delete": [<func>]
        # }

        for name in dir(self):
            o = getattr(self, name)
            if issubclass(o.__class__, EventListenerInherit):  # check for command objects
                if o.event not in listeners:
                    listeners[o.event] = []

                listeners[o.event].append(o)

                self.log.debug("found event listener")

        return listeners

    def find_bg_tasks(self):
        """Returns dict of background tasks."""

        self.log.debug("finding background tasks in {!r}".format(self))

        tasks = {}

        for name in dir(self):
            o = getattr(self, name)
            if issubclass(o.__class__, BgTaskInherit):  # check for command objects
                tasks[o.id] = o

                self.log.debug("found background task: {!r}".format(o.id))

        return tasks

    def create_task(self, *args, **kwargs):
        """
        Shortcut to :meth:`Plugin.bot.loop.create_task`

        Call this on a coroutine to run it without blocking.

        :returns: asyncio.Task
        """

        return self.bot.loop.create_task(*args, **kwargs)

    # event pre-processing, command handling

    async def __on_event__(self, event, *args, **kwargs):
        # triggers event listeners

        for listener in self.event_listeners.get(event, []):  # empty list if no event listeners
            self.create_task(listener.execute(self, *args, **kwargs))  # use plugin as self arg

            self.log.debug("{!r} event listener triggered".format(event))

    async def __on_ready__(self):
        for task in self.bg_tasks.values():
            task.restart(self.bot.loop, self)

    async def __on_shard_ready__(self):
        pass

    async def __on_message__(self, message):
        pass



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import discord
import aiohttp

from detache.command import Context
from detache import errors

import inspect

default_log = logging.getLogger("outlet")


class Bot(discord.Client):
    """
    Bot class. All of the code for your bot is written in plugins then registered with the bot.

    :keyword str default_prefix: (Optional) Default bot prefix. This can be overrided for per-server prefixes.
    :keyword logger: Logging object. The Detache log is used by default.
    """

    def __init__(self, *, default_prefix="!", logger=default_log):
        super().__init__()

        self.log = logger

        self.http_session = aiohttp.ClientSession()

        self.default_prefix = default_prefix

        async def get_prefix(guild):
            return self.default_prefix

        self.get_prefix = get_prefix

        self.plugins = []

        self.commands = {}

    def register_plugin(self, plugin, name=None):
        """
        Registers plugin to the bot.

        :param plugin: Plugin class. Must inherit :class:`detache.Plugin`
        :param str name: (Optional) Name of the plugin.
        """

        plugin = plugin(self)  # init plugin

        self.plugins.append(plugin)  # add to list
        self.commands.update(**plugin.commands)  # add commands to dict

    def plugin(self, name=None):
        """
        Plugin decorator for use in single file bots. Put this decorator before a plugin class for it to be registered
        automatically.

        :param str name: (Optional) Name of the plugin.
        """

        def decorator(plugin):
            self.register_plugin(plugin, name=name)

        return decorator

    def prefix(self, func):
        """
        To use a different command prefix for every guild your bot is in, put this decorator on a function that takes a
        :class:`discord.Guild` and returns the bot prefix. For example ::

            bot = detache.Bot()

            @bot.prefix()
            def get_prefix(ctx):
                # search for prefix in database
                # if not found, fallback to "!"

                return Prefixes.filter(guild_id=ctx.guild.id).first() or "!"

        If the get_prefix function returns None, the default prefix will be used instead.

        This decorator can be used on functions or coroutines.
        """

        async def get_prefix(guild):
            # use await if the get_prefix function is a coroutine
            if inspect.iscoroutine(func) or inspect.iscoroutinefunction(func):
                return await func(guild) or self.default_prefix
            else:
                return func(guild) or self.default_prefix

        # update with new func
        self.get_prefix = get_prefix

    # event handling

    async def on_message(self, message):
        if isinstance(message.channel, discord.TextChannel) and message.author != self.user:
            # happened in a guild, could be a command

            prefix = await self.get_prefix(message.guild)

            content = message.content
            if content.startswith(prefix) and content != prefix:
                self.log.debug("command called: {!r}".format(message.content))

                split = content[len(prefix):].split(" ")  # remove prefix, split up args and command

                cmd = split[0]
                args = " ".join(split[1:]) if len(split) > 0 else ""  # put args back together

                # check if command exists
                if cmd in self.commands:
                    command_object = self.commands[cmd]

                    # create command context
                    ctx = Context(command_object.plugin, message, prefix)

                    # attempt command
                    try:
                        await command_object.process(ctx, args)
                    except errors.CommandError as e:  # parsing error, i.e. wrong arg type
                        await message.channel.send(e)
                else:
                    # command does not exist!!
                    await message.channel.send("{}**{}** isn't a command.".format(prefix, cmd))

        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_message", message))
            self.loop.create_task(plugin.__on_message__(message))

    async def on_ready(self):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_ready"))
            self.loop.create_task(plugin.__on_ready__())

    async def on_shard_ready(self):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on__shard_ready"))
            self.loop.create_task(plugin.__on_shard_ready__())

    # passthrough

    async def on_typing(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_typing", *args, **kwargs))

    # messages

    async def on_message_delete(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_message_delete", *args, **kwargs))

    async def on_raw_message_delete(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_raw_message_delete", *args, **kwargs))

    async def on_message_edit(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_message_edit", *args, **kwargs))

    # reactions

    async def on_reaction_add(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_reaction_add", *args, **kwargs))

    async def on_reaction_remove(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_reaction_remove", *args, **kwargs))

    async def on_reaction_clear(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_reaction_clear", *args, **kwargs))

    # private channels

    async def on_private_channel_delete(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_private_channel_delete", *args, **kwargs))

    async def on_private_channel_create(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_private_channel_create", *args, **kwargs))

    async def on_private_channel_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_private_channel_update", *args, **kwargs))

    # guild channels

    async def on_guild_channel_delete(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_channel_delete", *args, **kwargs))

    async def on_guild_channel_create(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_channel_create", *args, **kwargs))

    async def on_guild_channel_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_channel_update", *args, **kwargs))

    # members

    async def on_member_join(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_member_join", *args, **kwargs))

    async def on_member_remove(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_member_remove", *args, **kwargs))

    async def on_member_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_member_update", *args, **kwargs))

    async def on_member_ban(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_member_ban", *args, **kwargs))

    async def on_member_unban(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_member_unban", *args, **kwargs))

    # guilds

    async def on_guild_join(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_join", *args, **kwargs))

    async def on_guild_remove(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_remove", *args, **kwargs))

    async def on_guild_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_update", *args, **kwargs))

    # roles

    async def on_guild_role_create(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_role_create", *args, **kwargs))

    async def on_guild_role_delete(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_role_delete", *args, **kwargs))

    async def on_guild_role_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_role_update", *args, **kwargs))

    # emojis

    async def on_guild_emojis_update(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_emojis_update", *args, **kwargs))

    # guild availability

    async def on_guild_available(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_available", *args, **kwargs))

    async def on_guild_unavailable(self, *args, **kwargs):
        for plugin in self.plugins:
            self.loop.create_task(plugin.__on_event__("on_guild_unavailable", *args, **kwargs))



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# used for detection by plugin
class EventListenerInherit:
    pass


def event_listener(event, one_task=False):
    """
    Event listener decorator. Calls function every time the given event occurs, with the respective arguments.

    :param event: Event to listen for.
    :param bool one_task: (Optional) If True, this will check if the event listener is already running when it's called.
                          If so, the running task is cancelled.
    """

    # class wraps the callback function
    # class is detected by plugin and plugin triggers the event listener
    class EventListener(EventListenerInherit):
        def __init__(self, func):
            self.func = func
            self.event = event

            self.one_task = one_task

            self.task = None

        async def execute(self, *args, **kwargs):
            if self.one_task and self.task is not None and not self.task.done():  # check if old task still runnning
                self.task.cancel()

            self.task = await self.func(*args, **kwargs)

    return EventListener


# used for detection by plugin
class BgTaskInherit:
    pass


def background_task(id):
    """
    Background task decorator. Runs a coroutine in the background when the bot connects to Discord.

    :param id: Arbitrary id for the background task.
    """

    # wrapper class
    class BgTask(BgTaskInherit):
        def __init__(self, func):
            self.id = id

            self.func = func
            self.task = None

        def start(self, loop, self_):
            self.task = loop.create_task(self.func(self_))

        def cancel(self):
            if self.task is not None and not self.task.done():
                self.task.cancel()

        def restart(self, loop, self_):
            self.cancel()
            self.start(loop, self_)

    return BgTask



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import re

import discord

from detache import errors


class Context(object):
    """
    Command context, passed to command functions for easier handling

    :attr discord.Message message: Message
    :attr discord.Guild guild: Guild the message was sent in
    :attr discord.Channel channel: Channel the message was sent in
    :attr discord.Member: author: Author of the message
    """

    def __init__(self, plugin, message, prefix=""):
        self.plugin = plugin

        self.message = message

        self.guild = message.guild
        self.channel = message.channel
        self.author = message.author

        self.prefix = prefix

    async def send(self, *args, **kwargs):
        """
        Coroutine

        Sends a message in the context's channel. Pass the same arguments or keywords that you would to
        :meth:`discord.Channel.send`
        """

        await self.channel.send(*args, **kwargs)


# empty class returned when an type doesn't match
class NoMatch:
    pass


# argument types

class Any:
    pattern = "[^ ]+"

    @classmethod
    def convert(cls, ctx, raw):
        """
        Converts string argument to specified type.

        :param ctx: Context.
        :param str raw: Raw passed argument.
        """

        return raw

    @classmethod
    def consume(cls, ctx, args):
        """
        Parses an argument from an argument string, and returns the argument string with this argument consumed.

        :return: parsed, argString
        """

        match = re.match(cls.pattern, args, flags=re.IGNORECASE)

        if match:
            match = match[0]

            parsed = cls.convert(ctx, match)

            span = len(match)
            if span >= len(args):
                return parsed, ""

            return parsed, args[span + 1:]  # remove this argument from arguments string

        else:
            return NoMatch, args


class String(Any):
    pattern = r'("[^\n]+"|[^ \n]+)'

    @classmethod
    def convert(cls, ctx, raw):
        # if the string is multi-word, it is surrounded in quotes. remove them
        if raw[0] == raw[-1] == '"':
            return raw[1:-1]

        return raw


class Number(Any):
    pattern = "\d+(\.\d+)?"

    @classmethod
    def convert(cls, ctx, raw):
        try:
            if "." in raw:
                return float(raw)
            else:
                return int(raw)

        except ValueError:
            raise errors.WrongType("{!r} is not a number.".format(raw))


class User(Any):
    pattern = "(<@!?([0-9]+)>|.{2,32}#[0-9]{4})"

    @classmethod
    def convert(cls, ctx, raw):
        # if contains "#", user tag was passed. otherwise, mention
        if "#" in raw:
            member = ctx.guild.get_member_named(raw)
        else:
            user_id = int(re.search("[0-9]+", raw)[0])
            member = ctx.guild.get_member(user_id)

        if member is None:
            raise errors.ParsingError("{} isn't a member of {}.".format(raw, ctx.guild))

        return member


class Channel(Any):
    pattern = r'(<#([0-9]+)>|#.{1,255})'

    @classmethod
    def convert(cls, ctx, raw):
        # if starts with "#", name of channel was passed.
        if raw.startswith("#"):
            name = raw[1:]

            channel = discord.utils.get(ctx.guild.text_channels, name=name)
        else:
            channel_id = int(re.search("[0-9]+", raw)[0])

            channel = discord.utils.get(ctx.guild.text_channels, id=channel_id)

        if channel is None:
            raise errors.ParsingError("{} isn't a channel in {}.".format(raw, ctx.guild))

        return channel


class Role(Any):
    pattern = r'(<@&[0-9]+>|"[^\n]+"|[^ \n]+)'

    @classmethod
    def convert(cls, ctx, raw):
        # if starts with "<@&", role mention was passed.
        if raw.startswith("<@&") and raw.endswith(">"):
            role_id = int(re.search("[0-9]+", raw)[0])

            role = discord.utils.get(ctx.guild.roles, id=role_id)
        else:
            name = raw

            if name[0] == name[-1] == '"':  # multi word, remove quotes
                name = name[1:-1]

            role = discord.utils.get(ctx.guild.roles, name=name)

        if role is None:
            raise errors.ParsingError("{} isn't a role in {}.".format(raw, ctx.guild))

        return role


# argument decorator

def argument(name, type=None, default=None, required=True, nargs=1, help=None):
    """
    Command argument.

    :param name: Name of the argument. Should match one of the function's arguments.
    :param type: (Optional) Argument type. Leave as None to accept any type.
    :param default: (Optional) Default value.
    :param required: (Optional) Whether the argument is required. Defaults to True
    :param nargs: (Optional) Number of times the argument can occur. Defaults to 1. -1 allows unlimited arguments.
    :param help: (Optional) Argument description.

    If nargs is anything other than 1, the parsed argument will be returned as a list.
    """

    if nargs not in (1, -1) and not required:
        raise ValueError("nargs must = 1 or -1 for arguments that aren't required")

    type = type or Any  # default ArgumentType class accepts anything as valid argument

    class Argument:
        @classmethod
        def no_match_error(cls):
            if nargs == 1:
                raise errors.ParsingError(
                    "**{}** is a required {}.".format(name, type.__name__.lower())
                )
            else:
                raise errors.ParsingError(
                    "**{}** are required.".format(name + ("" if name.endswith("s") else "s"))  # use plural
                )

        @classmethod
        def consume(cls, ctx, args):
            """
            Parses an argument from an argument string, and returns the argument string with this argument consumed.

            :return: parsed, argString
            """

            parsed, args = type.consume(ctx, args)  # use argument type's parsing function

            if parsed is NoMatch:  # argument is wrong type or not found
                if required or nargs != 1:
                    cls.no_match_error()
                else:
                    parsed = default

            return parsed, args

    Argument.name = name
    Argument.help = help
    Argument.type_ = type
    Argument.nargs = nargs
    Argument.required = required

    # actual decorator
    def add_argument(func):
        if hasattr(func, "cmd_args"):
            func.cmd_args.append(Argument)  # add to command function's arg list
        else:
            func.cmd_args = [Argument]  # arg list doesnt exist, create it

        return func

    return add_argument


# used to check plugin for commands
class CommandInherit:
    pass


def command(name, description=None, required_permissions=None):
    """
    Command decorator. Put this before a command and its arguments.

    :param str name: Name of command
    :param str description: Description of commands
    :param list[str] required_permissions: (Optional) Permissions required to use command
    """

    class Command(CommandInherit):
        def __init__(self, func):
            self.name = name
            self.description = description or inspect.cleandoc(inspect.getdoc(func))

            self.args = list(reversed(getattr(func, "cmd_args", [])))  # fix order of arguments

            self.func = func

            self.__doc__ = self.make_doc()

        def __repr__(self):
            return "Command({!r})".format(self.name)

        def make_doc(self, prefix=""):
            doc = "{}**{}** ".format(prefix, self.name) + " ".join([arg.name for arg in self.args]) + "\n\n"  # syntax

            # list arg types, names, descriptions
            for arg in self.args:
                arg_name = arg.type_.__name__ + ("(s)" if arg.nargs != 1 else "")

                doc += "• {} **{}**".format(arg_name, arg.name)

                if arg.help is not None:
                    doc += " - {}".format(arg.help)

                doc += "\n"

            doc += "\n" + self.description

            return doc

        async def process(self, ctx, content):
            # process given arguments and run the command

            # check for required permissions before parsing
            if required_permissions is not None:
                author_perms = ctx.author.permissions_in(ctx.channel)

                for perm in required_permissions:
                    # check permission. if not specified in permissions, assume False
                    if not getattr(author_perms, perm, False):
                        raise errors.MissingPermissions("This command requires the `{}` permission.".format(perm))

            parsed_args = {}

            try:
                for arg in self.args:
                    # parse argument and update with what's left of argument string

                    if arg.nargs == 1:  # only 1 arg
                        parsed, content = arg.consume(ctx, content)

                        parsed_args[arg.name] = parsed

                    elif arg.nargs == -1:  # any number of args
                        parsed = []

                        while True:
                            try:
                                value, content = arg.consume(ctx, content)

                                parsed.append(value)
                            except errors.ParsingError as e:  # no more args
                                if len(parsed) == 0 and arg.required:
                                    # must pass at least one arg if it's required
                                    raise e

                                break

                        parsed_args[arg.name] = parsed
                    else:
                        parsed = []

                        for i in range(arg.nargs):  # limit to nargs
                            try:
                                value, content = arg.consume(ctx, content)

                                parsed.append(value)
                            except errors.ParsingError:  # no more args
                                break

                        parsed_args[arg.name] = parsed

            except errors.ParsingError as e:
                raise errors.ParsingError("{}\n\n{}".format(e, self.make_doc(ctx.prefix)))

            reply = await self.func(ctx.plugin, ctx, **parsed_args)

            if reply:
                await ctx.send(reply)

    return Command



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import asyncio
import importlib.util
import os
from math import ceil as _ceil

import discord

ceil = lambda x: int(_ceil(x))


def import_file(path):
    try:
        spec = importlib.util.spec_from_file_location(os.path.basename(path), path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        return module
    except Exception as e:
        raise ImportError(e)


async def wait_then(s, coroutine):
    """
    Waits for s seconds before executing coroutine.

    :param str s: Seconds to wait
    :param asyncio.Coroutine coroutine: Coroutine
    :return: Return value of coroutine
    """
    await asyncio.sleep(s)

    return await coroutine


class PagedEmbed(object):
    """
    Utility class used to create paged embeds, which use reactions to go to other pages.

    :param discord.Embed embed: Embed to page. You should create your embed beforehand, this class
                                will simply paginate it.
    :param discord.Client client: Discord client to use when sending the paged embed.
    :param int per_page: (Keyword) Number of fields to show per page. Must be <= 25
    """

    __slots__ = ["_client", "_embed", "per_page", "author", "title", "description", "fields", "footer",
                 "pages", "current_page"]

    def __init__(self, client, embed, *, per_page=10):
        if not 1 <= per_page <= 25:
            raise ValueError("fields per page must be in range [1, 25]")

        self._client = client
        self._embed = embed

        self.per_page = per_page

        self.author = embed.author

        self.title = embed.title
        self.description = embed.description

        self.fields = embed.fields

        self.footer = embed.footer

        self.pages = self.make_pages()

        self.current_page = 0

    def make_pages(self):
        if len(self.fields) < self.per_page:
            return [self._embed]
        else:
            n_pages = ceil(len(self.fields) / self.per_page)

            embeds = []
            for page in range(n_pages):
                this_page = self.fields[:self.per_page]

                page_embed = discord.Embed(title=self.title, description=self.description)

                if self.author.name is not discord.Embed.Empty:
                    page_embed.set_author(**self.author)

                footer_text = self.footer.text + " | " if self.footer.text is not discord.Embed.Empty else ""
                page_embed.set_footer(text=footer_text + "Page {}/{}".format(page + 1, n_pages),
                                      icon_url=self.footer.icon_url)

                for field in this_page:
                    page_embed.add_field(name=field.name, value=field.value, inline=False)

                embeds.append(page_embed)

                del self.fields[:self.per_page]

            return embeds

    async def next_page(self, message):
        if self.current_page < len(self.pages):
            self.current_page += 1
            await message.edit(embed=self.pages[self.current_page])

    async def previous_page(self, message):
        if self.current_page > 0:
            self.current_page -= 1
            await message.edit(embed=self.pages[self.current_page])

    reactions = {
        "⬅": previous_page,
        "➡": next_page,
    }

    async def run(self, channel, user_for, *, timeout=120):
        """
        Runs the paged embed in specified channel.

        :param channel: Channel to run the embed in.
        :param user_for: Which user can use the page buttons.
        :param timeout: (Keyword) If no reaction is added in timeout seconds, the embed will stop responding."""

        def check(reaction_, user_):
            if user_ == self._client.user:  # bot added reaction
                return False

            if str(reaction_.emoji) not in self.reactions:  # check if reaction has any function
                return False

            return True

        message = await channel.send(embed=self.pages[self.current_page])

        for emoji in self.reactions:
            await message.add_reaction(emoji)

        while True:
            try:
                reaction, user = await self._client.wait_for("reaction_add", check=check, timeout=timeout)

                if user == user_for:
                    await self.reactions[str(reaction.emoji)](self, message)

                await message.remove_reaction(reaction, user)

            except asyncio.TimeoutError:
                await message.clear_reactions()
                break
